#!/usr/bin/env python3
#
# Copyright (c), 2015-2019, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
# Authors: Davide Brunato
#
"""
Test classes for Quantum Espresso input converters.
"""
import unittest
import os
import glob
import platform
import xml.etree.ElementTree as ElementTree

try:
    import yaml
except ImportError:
    yaml = None

import qeschema


def make_test_function(xml_file, ref_in_file):
    def test(self):
        tree = ElementTree.parse(xml_file)
        root = tree.getroot()

        element_name = root.tag.split('}')[-1]

        if element_name == 'espresso':
            xml_conf = qeschema.PwDocument(source=xml_file)
        elif element_name == 'nebRun':
            xml_conf = qeschema.NebDocument(source=xml_file, schema='qes_neb_test_ref.xsd')
        elif element_name == 'espressoph':
            xml_conf = qeschema.PhononDocument(source=xml_file)
        elif element_name == 'tddfpt':
            xml_conf = qeschema.TdDocument(source=xml_file)
        elif element_name == 'spectrumDoc':
            xml_conf = qeschema.TdSpectrumDocument(source=xml_file)
        elif element_name == 'xspectra':
            xml_conf = qeschema.XSpectraDocument(source=xml_file)
        elif element_name == 'epw':
            xml_conf = qeschema.EPWDocument(source=xml_file)
        else:
            raise ValueError("XML file %r is not a Quantum ESPRESSO document!" % xml_file)

        xml_conf.read(xml_file)
        qe_input = xml_conf.get_fortran_input().split('\n')
        with open(ref_in_file, 'r') as qe_input_file:
            k = 0
            are_equals = True
            for ref_line in qe_input_file:
                ref_line = ref_line.rstrip('\n').strip(' \t')
                in_line = qe_input[k].strip(' \t')
                if ref_line != in_line:
                    print("Unmatched lines: '%s' != '%s'" % (in_line, ref_line))
                    are_equals = False
                    break
                else:
                    k += 1
        self.assertTrue(are_equals, xml_file)
    return test


class ConverterTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.test_dir = os.path.dirname(os.path.abspath(__file__))
        cls.pkg_folder = os.path.dirname(cls.test_dir)

    def test_xml2qeinput_script(self):
        xml_filename = os.path.join(self.test_dir, 'resources/pw/Al001_relax_bfgs.xml')
        in_filename = xml_filename[:-4] + '.in'
        py_filename = os.path.join(self.pkg_folder, 'scripts/xml2qeinput.py')
        if os.path.isfile(in_filename):
            os.remove(in_filename)

        if platform.system() == 'Windows':
            os.system("python %s -in %s" % (py_filename, xml_filename))
            self.assertTrue('Al001_relax_bfgs.in' in os.listdir(os.path.dirname(xml_filename)))
        else:
            if platform.system() == 'Linux':
                command = 'python %s -in %s 1> /dev/null 2> /dev/null'
            else:
                command = 'python3 %s -in %s 1> /dev/null 2> /dev/null'

            os.system(command % (py_filename, xml_filename))
            self.assertTrue(os.path.isfile(in_filename),
                            'Test output file %r missing!' % in_filename)

    @unittest.skipIf(yaml is None, "PyYAML library is not installed")
    def test_yaml2qeinput_script(self):
        xml_filename = os.path.join(self.test_dir, 'resources/pw/Al001_relax_bfgs.yml')
        in_filename = xml_filename[:-4] + '.in'
        py_filename = os.path.join(self.pkg_folder, 'scripts/yaml2qeinput.py')
        if os.path.isfile(in_filename):
            os.remove(in_filename)

        if platform.system() == 'Windows':
            os.system("python %s -schema qes_test_ref.xsd -in %s" % (py_filename, xml_filename))
            self.assertTrue('Al001_relax_bfgs.in' in os.listdir(os.path.dirname(xml_filename)))
        else:
            if platform.system() == 'Linux':
                command = 'python %s -schema qes_test_ref.xsd -in %s 1> /dev/null 2> /dev/null'
            else:
                command = 'python3 %s -schema qes_test_ref.xsd -in %s 1> /dev/null 2> /dev/null'

            os.system(command % (py_filename, xml_filename))
            self.assertTrue(os.path.isfile(in_filename),
                            'Test output file %r missing!' % in_filename)


##
# Create test classes for examples
#
test_dir = os.path.dirname(os.path.abspath(__file__))

for filename in glob.glob(os.path.join(test_dir, "resources/*/*.xml")):
    qe_input_filename = '%s.in.test' % filename[:-4]
    if not os.path.isfile(qe_input_filename):
        continue

    test_func = make_test_function(filename, qe_input_filename)
    test_name = os.path.relpath(filename)
    klassname = 'Test_{0}'.format(test_name.replace('/', '__'))
    globals()[klassname] = type(
        klassname, (unittest.TestCase,),
        {'test_converter_{0}'.format(test_name): test_func, 'longMessage': True}
    )


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c), 2015-2020, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
# Authors: Davide Brunato
#
import logging
from collections.abc import MutableMapping

logger = logging.getLogger('qeschema')


def set_logger(loglevel=1, logfile=None):
    """
    Setup a basic logger with an handler and a formatter, using a
    corresponding numerical range [0..4], where a higher value means
    a more verbose logging. The loglevel value is mapped to correspondent
    logging module's value:

    LOG_CRIT=0 (syslog.h value is 2) ==> logging.CRITICAL
    LOG_ERR=1 (syslog.h value is 3) ==> logging.ERROR
    LOG_WARNING=2 (syslog.h value is 4) ==> logging.WARNING
    LOG_INFO=3 (syslog.h value is 6) ==> logging.INFO
    LOG_DEBUG=4 (syslog.h value is 7) ==> logging.DEBUG

    If a logfile name is passed then writes logs to file, instead of
    send logs to the standard output.

    :param loglevel: Simplified POSIX's syslog like logging level index
    :param logfile: Logfile name for non-scripts runs
    """
    global logger

    # Higher or lesser argument values are also mapped to DEBUG or CRITICAL
    effective_level = max(logging.DEBUG, logging.CRITICAL - loglevel * 10)

    logger.setLevel(effective_level)

    # Add the first new handler
    if not logger.handlers:
        if logfile is None:
            lh = logging.StreamHandler()
        else:
            lh = logging.FileHandler(logfile)
        lh.setLevel(effective_level)

        if effective_level <= logging.DEBUG:
            fmt = "[%(levelname)s:%(module)s:%(funcName)s: %(lineno)s] %(message)s"
        elif effective_level <= logging.INFO:
            fmt = "[%(levelname)s:%(module)s] %(message)s"
        else:
            fmt = "%(levelname)s: %(message)s"

        formatter = logging.Formatter(fmt)
        lh.setFormatter(formatter)
        logger.addHandler(lh)
    else:
        for handler in logger.handlers:
            handler.setLevel(effective_level)


def etree_iter_path(elem, tag=None, path='.'):
    """
    Iterate an ElementTree structure giving back couples with an element and its path.

    :param elem: root of the ElementTree
    :param tag: Optional tag matching argument.
    :param path: Argument element path
    """
    if tag == "*":
        tag = None
    if tag is None or elem.tag == tag:
        yield elem, path
    for child in elem:
        child_path = '%s/%s' % (path, child.tag)
        for e, p in etree_iter_path(child, tag, path=child_path):
            yield e, p


def to_fortran(value):
    """
    Translate a Python value to the equivalent literal representation for Fortran input.
    Leading and trailing spaces characters are removed from strings.
    """
    if isinstance(value, bool):
        return '.true.' if value else '.false.'
    elif isinstance(value, str):
        return repr(value.strip())
    elif isinstance(value, bytes):
        return repr(value.decode('utf-8').strip())
    return str(value)


def get_target_prefix(namespaces, target_namespace):
    """Returns the longest prefix that maps the target namespace."""
    prefix = None
    for k, v in namespaces.items():
        if v == target_namespace:
            if prefix is None or len(k) > len(prefix):
                prefix = k
    else:
        return prefix


class BiunivocalMap(MutableMapping):
    """
    A dictionary that implements a bijective correspondence, namely with constraints
    of uniqueness both on keys and on values. Both keys and values must be hashable.
    """
    def __init__(self, *args, **kwargs):
        self.__map = {}
        self.__inverse = {}
        self.update(*args, **kwargs)

    def __getitem__(self, key):
        try:
            return self.__map[key]
        except KeyError:
            if not hasattr(self, '__missing__'):
                raise
            return getattr(self, '__missing__')(key)

    def __setitem__(self, key, item):
        try:
            if self.__inverse[item] != key:
                raise ValueError("Value '{0}' is already mapped by another key!".format(item))
        except KeyError:
            if key in self.__map:
                del self.__inverse[self.__map[key]]
            self.__map[key] = item
            self.__inverse[item] = key
        else:
            del self.__inverse[self.__map[key]]
            self.__map[key] = item
            self.__inverse[item] = key

    def __delitem__(self, key):
        del self.__inverse[self.__map[key]]
        del self.__map[key]

    def __iter__(self):
        return iter(self.__map)

    def __len__(self):
        return len(self.__map)

    def __contains__(self, key):
        return key in self.__map

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.__map or '')

    def copy(self):
        return self.__class__(self.__map.copy())

    def inverse(self):
        """Returns a copy of the inverse dictionary."""
        return self.__inverse.copy()

    def getkey(self, value, default=None):
        """
        If *value* is in dictionary's values, returns the key correspondent to the value,
        else returns the *default* argument.
        """
        try:
            return self.__inverse[value]
        except KeyError:
            return default



#
# Copyright (c), 2015-2020, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
# Authors: Davide Brunato
#
import logging
import os.path
import json
from abc import ABCMeta
from functools import wraps
from xml.etree import ElementTree
import xmlschema
from xmlschema import etree_tostring

try:
    import yaml
except ImportError:
    yaml = None

from .namespaces import XSD_NAMESPACE
from .converters import RawInputConverter, PwInputConverter, PhononInputConverter, \
    NebInputConverter, TdInputConverter, TdSpectrumInputConverter, \
    XSpectraInputConverter, EPWInputConverter
from .exceptions import XmlDocumentError
from .utils import etree_iter_path, get_target_prefix

logger = logging.getLogger('qeschema')

SCHEMAS_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'schemas')


def requires_xml_data(method):
    """A decorator for XML document methods that require XML data to be loaded."""
    @wraps(method)
    def check_xml_data(self, *args, **kwargs):
        if self.root is None:
            raise XmlDocumentError("No XML data loaded!")
        return method(self, *args, **kwargs)
    return check_xml_data


def removeprefix(s, prefix):
    return s[len(prefix):] if s.startswith(prefix) else s


class XmlDocument:
    """
    Base class for a generic XML document based on an XSD schema. The schema
    associated is used for checking types, validation of the XML data and for
    lookup of default values. XML data is loaded into memory into an ElementTree
    structure. Data files can be also in JSON or YAML format, in these cases the
    data source is converted to XML when loading.

    :param source: can be a :class:`xmlschema.XMLResource` instance or a file-like \
    object or a file path or an URL of a resource or a string containing the XML data.
    :param schema: can be a :class:`xmlschema.XMLSchema` instance or a file-like \
    object or a file path or an URL of a resource or a string containing the XSD schema.

    :cvar SEARCH_PATHS: the sequence of search paths used by :meth:`fetch_schema` \
    for fetching schemas.
    :ivar root: the root element of the XML tree.
    :ivar filename: the filepath of the data source file.
    :ivar format: the format of the data source file (XML, JSON, YAML).
    :ivar errors: the list of detected validation errors.
    :ivar schema: the :class:`XMLSchema` instance associated with the document.
    """
    SEARCH_PATHS = ('.',)
    DEFAULT_SCHEMA = None

    def __init__(self, source=None, schema=None):
        self.root = None
        self.filename = None
        self.format = None
        self.errors = []
        self._namespaces = {}

        if source is None:
            source_schema = None
        else:
            if not isinstance(source, xmlschema.XMLResource):
                source = xmlschema.XMLResource(source)

            if source.namespace == XSD_NAMESPACE:
                raise XmlDocumentError("source is an XSD schema")

            for ns, location in source.iter_location_hints():
                if ns == source.namespace:
                    source_schema = self.fetch_schema(location)
                    if source_schema is not None:
                        break
            else:
                source_schema = None

        if isinstance(schema, xmlschema.XMLSchemaBase):
            self.schema = schema
        elif isinstance(schema, str) and '\n' not in schema \
                and not schema.lstrip().startswith('<'):
            self.schema = xmlschema.XMLSchema(self.fetch_schema(schema) or schema)
        elif schema is not None:
            self.schema = xmlschema.XMLSchema(schema)
        elif source_schema is not None:
            self.schema = xmlschema.XMLSchema(source_schema)
        elif self.DEFAULT_SCHEMA is not None:
            default_schema = self.fetch_schema(self.DEFAULT_SCHEMA)
            self.schema = xmlschema.XMLSchema(default_schema)
        else:
            raise XmlDocumentError("missing schema for XML data!")

        if source is not None:
            self.from_xml(source, validation='lax')

    @property
    def namespaces(self):
        """
        XML data namespaces map, a dictionary that maps prefixes to URI. An empty
        dictionary if the XML data file is not loaded or if it doesn't contain any
        namespace declaration.
        """
        return {k: v for k, v in self._namespaces.items()}

    @property
    def default_namespace(self):
        return self._namespaces.get('', '')

    @classmethod
    def fetch_schema(cls, filename):
        filename = filename.strip()
        if os.path.isfile(filename):
            return filename

        if not filename.startswith('/'):
            for base_path in cls.SEARCH_PATHS:
                if os.path.isfile(os.path.join(base_path, filename)):
                    return os.path.join(base_path, filename)

        base_name = os.path.basename(filename)
        for base_path in cls.SEARCH_PATHS:
            if os.path.isfile(os.path.join(base_path, base_name)):
                return os.path.join(base_path, base_name)

    def read(self, filename, validation='strict', **kwargs):
        """
        Reads XML data from a file encoded in XML, JSON or YAML format.

        :param filename: filepath of the data source file.
        :param validation: validation mode, can be 'strict', 'lax' or 'skip'.
        :param kwargs: other options to pass to the encoding method of the schema \
        instance in case of a non-XML data source.
        """
        if not isinstance(filename, str):
            raise TypeError("wrong type for argument 'filename'")
        elif not os.path.isfile(filename):
            raise ValueError("{!r} is not a file".format(filename))

        ext = filename.strip().lower().rpartition('.')[2] if '.' in filename else None

        if ext == 'xml':
            self.from_xml(filename, validation)
        elif ext == 'json':
            self.from_json(filename, validation, **kwargs)
        elif ext in ('yml', 'yaml'):
            self.from_yaml(filename, validation, **kwargs)
        else:
            try:
                self.from_xml(filename, validation)
            except (ElementTree.ParseError, SyntaxError):
                try:
                    self.from_json(filename, validation, **kwargs)
                except json.JSONDecodeError:
                    try:
                        self.from_yaml(filename, validation, **kwargs)
                    except yaml.YAMLError:
                        raise ValueError(
                            "input file is not in neither of XML, JSON or YAML formats"
                        )

    def from_xml(self, source, validation='strict', **kwargs):
        """
        Load XML data. Data is validated against the schema.

        :param source: a filepath to an XML file or a string containing XML data.
        :param validation: validation mode, can be 'strict', 'lax' or 'skip'.
        :param kwargs: other options for creating the :class:`xmlschema.XMLResource` \
        instance used for reading the XML data.
        :return: a couple with the root element of the XML ElementTree a list \
        containing the detected errors.
        """
        if not isinstance(source, xmlschema.XMLResource):
            source = xmlschema.XMLResource(source, **kwargs)

        errors = []
        if validation == 'strict':
            self.schema.validate(source)
        elif validation == 'lax':
            errors.extend(e for e in self.schema.iter_errors(source))

        self.root = source.root
        self.errors = errors
        self._namespaces = source.get_namespaces()

        if source.url is None:
            self.filename = None
            self.format = None
        else:
            self.filename = removeprefix(source.url, 'file://')
            self.format = 'xml'

    def from_json(self, source, validation='strict', **kwargs):
        """
        Load JSON encoded data. Data is converted to an XML ElementTree structure
        and validated against the schema.

        :param source: a filepath to a JSON file or a string containing JSON data.
        :param validation: validation mode, can be 'strict', 'lax' or 'skip'.
        :param kwargs: other options to pass to the encoding method of the schema instance.
        :return: the root element of the XML ElementTree data structure and a list \
        containing the detected errors.
        :raise: an :class:`xmlschema.XMLSchemaValidationError` if validation is strict \
        and at least an error is found.
        """
        if not isinstance(source, str):
            raise TypeError("the source argument must be a string!")

        preserve_root = kwargs.pop('preserve_root', True)
        try:
            json.loads(source)
        except ValueError:
            with open(source) as f:
                obj = xmlschema.from_json(f, self.schema, validation=validation,
                                          preserve_root=preserve_root)
            filename = source.strip()
        else:
            obj = xmlschema.from_json(source, self.schema, validation=validation,
                                      preserve_root=preserve_root)
            filename = None

        if isinstance(obj, tuple):
            self.root, self.errors = obj
        else:
            self.root, self.errors = obj, []
        self.filename = filename
        self.format = 'json' if filename else None

    def from_yaml(self, source, validation='strict', **kwargs):
        """
        Converts a YAML encoded file to an XML ElementTree structure.
        Data is validated against the schema during conversion.

        :param source: a filepath to a YAML file or a string containing YAML data.
        :param validation: validation mode, can be 'strict', 'lax' or 'skip'.
        :param kwargs: other options to pass to the encoding method of the schema instance.
        :return: a couple with the root element of the XML ElementTree and a list \
        containing the detected errors.
        """
        if yaml is None:
            raise RuntimeError("PyYAML library is not installed!")
        elif not isinstance(source, str):
            raise TypeError("the source argument must be a string!")
        elif '\n' not in source and not source.strip().startswith('<'):
            with open(source) as f:
                data = yaml.load(f, Loader=yaml.Loader)
            filename = source.strip()
        else:
            data = yaml.load(source, Loader=yaml.Loader)
            filename = None

        preserve_root = kwargs.pop('preserve_root', True)
        converter = kwargs.pop('converter', xmlschema.UnorderedConverter)
        if 'path' not in kwargs and isinstance(data, dict) and len(data) == 1:
            kwargs['path'] = list(data.keys())[0]

        obj = self.schema.encode(data, validation=validation, converter=converter,
                                 preserve_root=preserve_root, **kwargs)

        if isinstance(obj, tuple):
            self.root, self.errors = obj
        else:
            self.root, self.errors = obj, []
        self.filename = filename
        self.format = 'yaml' if filename else None

    def from_dict(self, data, validation='strict', **kwargs):
        """
        Converts a Python object to an XML ElementTree structure.
        Object data is validated against the schema during conversion.

        :param data: filepath of the data source file.
        :param validation: validation mode, can be 'strict', 'lax' or 'skip'.
        :param kwargs: other options to pass to the encoding method of the schema instance.
        :return: a couple with the root element of the XML ElementTree and a list \
        containing the detected errors.
        """
        preserve_root = kwargs.pop('preserve_root', True)
        obj = self.schema.encode(data, validation=validation, preserve_root=preserve_root, **kwargs)
        if isinstance(obj, tuple):
            self.root, self.errors = obj
        else:
            self.root, self.errors = obj, []
        self.filename = self.format = None

    @requires_xml_data
    def write(self, filename, output_format='xml', validation='strict', **kwargs):
        """
        Write loaded XML data to a file. Binds the document to saved file if
        it's not already bound to another file.

        :param filename: filepath of the destination file.
        :param output_format: the data format of the output file.
        :param validation: validation mode, can be 'strict', 'lax' or 'skip'.
        :param kwargs: other options for the decoding method of the schema instance.
        """
        if not isinstance(filename, str):
            raise TypeError("the filename argument must be a string!")

        output_format = output_format.strip().lower()
        if output_format == 'xml':
            with open(filename, 'w+') as f:
                f.write(etree_tostring(self.root))

        elif output_format == 'json':
            obj = self.to_dict(validation, **kwargs)
            with open(filename, 'w+') as f:
                return json.dump(obj, f, sort_keys=True, indent=4)

        elif output_format == 'yaml':
            if yaml is None:
                raise RuntimeError("PyYAML library is not installed!")

            obj = self.to_dict(validation, **kwargs)
            with open(filename, 'w+') as f:
                yaml.dump(obj, stream=f, default_flow_style=False)
        else:
            raise ValueError("Accepted output_format are 'xml', 'json' or 'yaml'!")

        if self.filename is None:
            self.filename = filename
            self.format = output_format

    @requires_xml_data
    def to_dict(self, validation='strict', **kwargs):
        """
        Converts loaded XML data to a nested dictionary.

        :param validation: validation mode, can be 'strict', 'lax' or 'skip'.
        :param kwargs: other options for the decoding method of the schema instance.
        :returns: a dictionary.
        """
        obj = self.schema.to_dict(
            source=self.root,
            validation=validation,
            namespaces=kwargs.get('namespaces') or self.namespaces,
            preserve_root=kwargs.pop('preserve_root', True),
            **kwargs
        )
        return obj[0] if isinstance(obj, tuple) else obj

    def to_json(self, filename=None, validation='strict', **kwargs):
        """
        Converts loaded XML data to a JSON string or file.

        :param filename: filepath of the destination file. Binds the document to \
        this file if it's not already bound to another file.
        :param validation: validation mode, can be 'strict', 'lax' or 'skip'.
        :param kwargs: other options for the decoding method of the schema instance.
        """
        data = self.to_dict(validation, **kwargs)
        if filename is None:
            return json.dumps(data, sort_keys=True, indent=4)

        with open(filename, mode='w+') as f:
            json.dump(data, f, sort_keys=True, indent=4)

        if filename is not None and self.filename is None:
            self.filename = filename
            self.format = 'json'

    def to_yaml(self, filename=None, validation='strict', **kwargs):
        """
        Converts loaded XML data to YAML string or file.

        :param filename: filepath of the destination file. Binds the document to \
        this file if it's not already bound to another file.
        :param validation: validation mode, can be 'strict', 'lax' or 'skip'.
        :param kwargs: other options for the decoding method of the schema instance.
        """
        if yaml is None:
            raise RuntimeError("PyYAML library is not installed!")

        data = self.to_dict(validation, **kwargs)
        if filename is None:
            return yaml.dump(data, default_flow_style=False)

        with open(filename, mode='w+') as f:
            yaml.dump(data, stream=f, default_flow_style=False)

        if filename is not None and self.filename is None:
            self.filename = filename
            self.format = 'yaml'

    # ElementTree API wrappers

    def iter(self, tag=None):
        return self.root.iter(tag)

    def find(self, path, namespaces=None):
        """
        Find first matching element by tag name or path.

        :param path: is a string having either an element tag or an XPath,
        :param namespaces: is an optional mapping from namespace prefix to full name.
        :return: the first matching element, or None if no element was found
        """
        if path[:1] == "/":
            path = "." + path
        return self.root.find(path, namespaces)

    def findall(self, path, namespaces=None):
        """
        Find all matching subelements by tag name or path.

        :param path: is a string having either an element tag or an XPath,
        :param namespaces: is an optional mapping from namespace prefix to full name.
        :return: the first matching element, or None if no element was found
        """
        if path[:1] == "/":
            path = "." + path
        return self.root.findall(path, namespaces)


class QeDocument(XmlDocument, metaclass=ABCMeta):
    """
    Abstract base class for schema-based XML documents of Quantum ESPRESSO applications.
    """
    SEARCH_PATHS = (SCHEMAS_DIR, os.path.join(SCHEMAS_DIR, 'releases'), '.')
    DEFAULT_INPUT_BUILDER = None

    def __init__(self, source=None, schema=None, input_builder=None):
        super(QeDocument, self).__init__(source, schema)

        if input_builder is None:
            self.input_builder = self.DEFAULT_INPUT_BUILDER
        elif not isinstance(input_builder, type) or \
                not issubclass(input_builder, RawInputConverter):
            raise XmlDocumentError(f"3rd argument must be a {RawInputConverter!r} subclass")
        else:
            self.input_builder = input_builder

        self._xpath_namespaces = {k: v for k, v in self.schema.namespaces.items() if k}

        prefix = get_target_prefix(self.schema.namespaces, self.schema.target_namespace)
        if prefix not in ('qes', 'neb', 'qes_ph', 'qes_lr',
                          'qes_spectrum', 'qes_xspectra', 'epw'):
            raise NotImplementedError(
                f"Converter not implemented for namespace {self.schema.target_namespace!r}"
            )

    def _xsd_find(self, path, xsd_element=None):
        if xsd_element is None:
            return self.schema.find(path, self._xpath_namespaces)
        else:
            return xsd_element.find(path, self._xpath_namespaces)

    @property
    def input_path(self):
        """The path to XML input section."""
        return 'input'

    @property
    def output_path(self):
        """The path to XML input section."""
        return 'output'

    def write_fortran_input(self, filename):
        """
        Converts the XML input data to a Fortran namelist input and writes it to a file.

        :param filename: the pathname of the file to use to save the Fortran namelist.
        """
        with open(filename, mode='w+') as f:
            f.write(self.get_fortran_input())

    @requires_xml_data
    def get_fortran_input(self, use_defaults=True):
        """
        Converts the XML input data to a Fortran namelist input.

        :param use_defaults: use the defaults of the XSD schema to fill missing values.
        :returns: a string.
        """
        qe_input = self.input_builder(xml_file=self.filename)
        input_path = './%s' % self.input_path

        input_root = self.find(input_path)
        if input_root is None:
            raise XmlDocumentError("Missing input {!r} in XML data!".format(input_path))

        for schema_root in self.schema.elements.values():
            if self._xsd_find(input_path, schema_root):
                break
        else:
            raise XmlDocumentError("Missing input element in XSD schema!")

        # Extract values from input's subtree of the XML document
        for elem, path in etree_iter_path(input_root, path=input_path):
            rel_path = path.replace(input_path, '.')
            xsd_element = self._xsd_find(path, schema_root)
            if xsd_element is None:
                logger.error("%r doesn't match any element!", path)
                continue
            else:
                value = xsd_element.decode(elem, use_defaults=use_defaults)
                if isinstance(value, str):
                    value = value.strip()
                node_dict = {elem.tag: value}
            logger.debug("Add input for node '{0}' with dict '{1}'".format(elem.tag, node_dict))

            # Convert attributes
            for attr_name, value in elem.attrib.items():
                logger.debug("Convert attribute '%s' of element '%s'" % (attr_name, path))
                path_key = '%s/@%s' % (rel_path, attr_name)
                if path_key not in qe_input:
                    logger.debug("Attribute's path '%s' not in converter!" % path_key)
                    continue
                qe_input.set_path(path_key, elem.tag, node_dict)

            logger.debug("Convert element '%s'" % path)
            path_key = '%s/$' % rel_path if xsd_element.attributes else rel_path
            if path_key not in qe_input:
                logger.debug("Element's path '%s' not in converter!" % path_key)
                continue
            qe_input.set_path(path_key, elem.tag, node_dict)

        return qe_input.get_qe_input()


class PwDocument(QeDocument):
    """
    Class to manage PW XML documents.
    """
    DEFAULT_SCHEMA = 'qes.xsd'
    DEFAULT_INPUT_BUILDER = PwInputConverter

    @requires_xml_data
    def get_atomic_positions(self):
        """
        Gets atomic symbols and atomic positions from XML output data.

        :return: the list of atomic symbols and a nested list containing the coordinates
        """
        path = './/output//atomic_positions'
        elem = self.find(path)
        if elem is not None:
            atomic_positions = self._xsd_find(path).decode(elem)
            atoms = atomic_positions.get('atom')
            if not isinstance(atoms, list):
                atoms = [atoms]
            symbols = [a['@name'] for a in atoms]
            positions = [a['$'] for a in atoms]
            return symbols, positions

    @requires_xml_data
    def get_cell_parameters(self):
        """
        Gets cell parameters from an XML output data.

        :return: a nested list containing the cell vectors in Bohr atomic units
        """
        path = './/output//cell'
        elem = self.find(path)
        if elem is not None:
            cell = self._xsd_find(path).decode(elem)
            return [cell['a1'], cell['a2'], cell['a3']]

    @requires_xml_data
    def get_stress(self):
        """
        Gets stress tensor from the XML output data, if present.

        :return: nested list containing the stress tensor in C order
        """
        path = './/output//stress'
        elem = self.find(path)
        if elem is not None:
            stress = self._xsd_find(path).decode(elem)
            try:
                stress = stress['$']
            except TypeError:
                pass
            return [stress[::3], stress[1::3], stress[2::3]]

    @requires_xml_data
    def get_forces(self):
        """
        Gets forces from the XML output data, if present.

        :return: the list of atomic symbols plus a nested list with the forces \
        in atomic units
        """
        path = './/output/forces'
        elem = self.find(path)
        if elem is not None:
            forces = self._xsd_find(path).decode(elem)
            path = './/output//atomic_positions'
            atomic_positions = self._xsd_find(path).decode(self.find(path))
            atoms = atomic_positions.get('atom', [])
            if not isinstance(atoms, list):
                atoms = [atoms]
            symbols = [a['@name'] for a in atoms]
            i0 = range(3 * len(atoms))[::3]
            i1 = range(3 * len(atoms) + 1)[3::3]
            forces = [forces['$'][i:j] for i, j in zip(i0, i1)]
            return symbols, forces

    @requires_xml_data
    def get_k_points(self):
        """
        Extracts the k_points list from the XML output data.

        :return: nested list with k_points
        """
        path = './/output//ks_energies/k_point'
        return [self._xsd_find(path).decode(e)['$'] for e in self.findall(path)]

    @requires_xml_data
    def get_ks_eigenvalues(self):
        """
        Extracts the eigenvalues from the XML output data.

        :return: nested list of KS eigenvalues for each k_point in Hartree Units
        """
        path = './/output//ks_energies/eigenvalues'
        eigenvalues = []
        for e in self.findall(path):
            obj = self._xsd_find(path).decode(e)
            if isinstance(obj, dict):
                eigenvalues.append(obj['$'])  # pragma: no cover
            else:
                eigenvalues.append(obj)

        return eigenvalues

    @requires_xml_data
    def get_total_energy(self):
        """
        Extracts the total energy from the  XML output data.

        :return: total energy in Hartree Units
        """
        path = './/output//etot'
        return self._xsd_find(path).decode(self.find(path))


class PhononDocument(QeDocument):
    """
    Class to manage Phonon XML documents.
    """
    DEFAULT_SCHEMA = 'ph_xmlschema.xsd'
    DEFAULT_INPUT_BUILDER = PhononInputConverter

    @property
    def input_path(self):
        return 'inputPH'

    @property
    def output_path(self):
        return 'outputPH'

    def get_fortran_input(self, use_defaults=False):
        """
        Overrides get_fortran_input() setting *use_defaults* optional argument to False.

        :param use_defaults:
        :return: the input as obtained from its input builder
        """
        return super(PhononDocument, self).get_fortran_input(use_defaults=use_defaults)


class NebDocument(QeDocument):
    """
    Class to manage NEB XML documents.
    """
    DEFAULT_SCHEMA = 'qes_neb.xsd'
    DEFAULT_INPUT_BUILDER = NebInputConverter


class TdDocument(QeDocument):
    """
    Class to manage TDDFPT XML documents.
    """
    DEFAULT_SCHEMA = 'tddfpt.xsd'
    DEFAULT_INPUT_BUILDER = TdInputConverter

    @property
    def input_path(self):
        return 'input'


class TdSpectrumDocument(QeDocument):
    """
    Class to manage turbo-spectrum XML inputs
    """
    DEFAULT_SCHEMA = 'qes_spectrum.xsd'
    DEFAULT_INPUT_BUILDER = TdSpectrumInputConverter

    @property
    def input_path(self):
        return 'spectrumIn'


# XSPECTRA custom document
class XSpectraDocument(QeDocument):
    """
    Class to manage XSPECTRA XML documents.
    """
    DEFAULT_SCHEMA = 'xspectra.xsd'
    DEFAULT_INPUT_BUILDER = XSpectraInputConverter

    @property
    def input_path(self):
        return 'input'


class EPWDocument(QeDocument):
    """
    class to manage EPW XML documents
    """
    DEFAULT_SCHEMA = 'epw.xsd'
    DEFAULT_INPUT_BUILDER = EPWInputConverter

    @property
    def input_path(self):
        return 'input'

    def get_fortran_input(self, use_defaults=False):
        """overrides get_fortran_input adding title-line on top and
           and setting *use_defaults* optional argument to false
        """
        path = './input/control_variables/title'
        element = self.find(path)
        if element is not None:
            title = str(element.text)
        else:
            title = "---"
        return "\n".join([title, super().get_fortran_input(use_defaults)])



#!/usr/bin/env python
#
# Copyright (c), 2015-2020, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
# Authors: Davide Brunato, Giovanni Borghi
#
"""
Convert from XML input to Fortran input
"""

import sys


def parse_args():
    """Command arguments parsing"""
    import argparse

    parser = argparse.ArgumentParser(
            description="This program converts an XML input to the an equivalent "
                        "input file written in a format that is natively readable "
                        "by Fortran's codes of Quantum Espresso"
    )
    parser.add_argument("-v", "--verbosity", action="count", default=1,
                        help="Increase output verbosity.")
    parser.add_argument('-in', metavar='FILE', required=True, help="XML input filename.")
    parser.add_argument('-schema', metavar='FILE', required=False,
                        help="Specify XSD schema in input", default=None)
    return parser.parse_args()


if __name__ == '__main__':

    if sys.version_info < (3, 7, 0):
        sys.stderr.write("You need python 3.7 or later to run this program\n")
        sys.exit(1)

    args = parse_args()
    print("Create Fortran input from XML file %r ...\n" % getattr(args, 'in'))

    if __package__ is None:
        from os import path
        sys.path.append(path.abspath(path.dirname(__file__)+'/../'))

    import qeschema
    import os
    import xml.etree.ElementTree as Etree

    qeschema.set_logger(args.verbosity)

    input_fn = getattr(args, 'in')
    schema_fn = getattr(args, 'schema', None)
    tree = Etree.parse(input_fn)
    root = tree.getroot()
    element_name = root.tag.split('}')[-1]
    if element_name == 'espresso':
        xml_document = qeschema.PwDocument(source=input_fn, schema=schema_fn)
    elif element_name == 'nebRun':
        xml_document = qeschema.NebDocument(source=input_fn, schema=schema_fn)
    elif element_name == 'espressoph':
        xml_document = qeschema.PhononDocument(source=input_fn, schema=schema_fn)
    elif element_name == 'tddfpt':
        xml_document = qeschema.TdDocument(source=input_fn, schema=schema_fn)
    elif element_name == 'spectrumDoc':
        xml_document = qeschema.TdSpectrumDocument(source=input_fn, schema=schema_fn)
    elif element_name == 'xspectra':
        xml_document = qeschema.XSpectraDocument(source=input_fn, schema=schema_fn)
    elif element_name == 'epw':
        xml_document = qeschema.EPWDocument(souce=input_fn, schema=schema_fn)
    else:
        sys.stderr.write("Could not find correct XML in %s, exiting...\n" % input_fn)
        sys.exit(1)

    root = None
    tree = None

    qe_in = xml_document.get_fortran_input()

    input_fn_name, input_fn_ext = os.path.splitext(input_fn)
    outfile = input_fn_name + '.in'

    with open(outfile, mode='w') as f:
        f.write(qe_in)
        print("Input configuration written to file '%s' ..." % outfile)



#
# Copyright (c), 2015-2020, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
# Authors: Davide Brunato
#
import logging

logger = logging.getLogger('qeschema')


class QESchemaError(Exception):
    pass


class XmlDocumentError(QESchemaError, RuntimeError):
    """An error or a wrong condition with an XML document instance."""

    def __init__(self, message):
        Exception.__init__(self, message)
        logger.debug('!XmlDocumentError: {0}'.format(message))



#
# Copyright (c), 2015-2020, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
# Authors: Davide Brunato, Giovanni Borghi
#
"""
Conversion functions for Quantum Espresso input options.
"""
import logging

from .exceptions import XmlDocumentError
from .utils import to_fortran

logger = logging.getLogger('qeschema')


#
# Other derived values
def get_specie_related_values(name, **kwargs):
    """
    Convert XML data for specie related options. Map single values,
    vectors and matrices. Skip 0 values for Hubbard parameters or
    negative values for starting_ns eigenvalues. Skip entire vector
    or matrix when

    :param name: parameter name
    :param kwargs: related input data
    :return: a list
    """
    related_tag = kwargs['_related_tag']
    related_data = kwargs[related_tag]
    try:
        new_format = kwargs['dftU']['@new_format']
    except KeyError:
        new_format = False

    if new_format and related_tag in ['Hubbard_U', 'Hubbard_J', 'Hubbard_alpha',
                                      'Hubbard_beta', 'Hubbard_J0']:
        return []
    try:
        atomic_species = kwargs['atomic_species']
        species = atomic_species['species']
    except KeyError as err:
        logger.error("Missing required argument %s when building "
                     "parameter %r", str(err), name)
        return []

    lines = []
    for value in iter(related_data if isinstance(related_data, list) else [related_data]):
        tag_specie = value['@specie']
        tag_values = value['$']
        tag_spin = value.get('@spin')
        if value.get('@label') == 'no Hubbard':
            continue

        specie_index = 1
        for specie in species:
            if specie['@name'] == tag_specie:
                break
            specie_index += 1
        else:
            raise XmlDocumentError("Unknown specie {!r} in tag {!r}".format(tag_specie, name))

        if isinstance(tag_values, list):
            for k in range(len(tag_values)):
                # starting_ns case: skip negative values
                if tag_values[k] < 0 or (name == 'Hubbard_J' and tag_values[k] == 0):
                    continue
                if tag_spin:
                    lines.append(' {0}({1},{2},{3})={4}'.format(
                        name, k + 1, tag_spin, specie_index, tag_values[k]
                    ))
                else:
                    lines.append(' {0}({1},{2})={3}'.format(
                        name, k + 1, specie_index, tag_values[k]
                    ))
        elif tag_values > 0:
            lines.append(' {0}({1})={2}'.format(name, specie_index, tag_values))
    return lines


def get_starting_magnetization(name, **kwargs):
    """
    Build starting magnetization vector from species data.
    """
    try:
        atomic_species = kwargs['atomic_species']
        species = atomic_species['species']
    except KeyError as err:
        logger.error("Missing required arguments when building "
                     "parameter '%s'! %s" % (name, err))
        return []

    lines = []
    try:
        lines.append(' {0}(1)={1}'.format(name, species.get('starting_magnetization', 0.0)))
    except AttributeError:
        k = 0
        for specie in species:
            k += 1
            lines.append(' {0}({1})={2}'.format(
                name, k, specie.get('starting_magnetization', 0.0)
            ))
    return lines


def get_system_nspin(name, **kwargs):
    """
    Get the value for 'nspin' parameter of the SYSTEM namelist.
    """
    try:
        lsda = kwargs['lsda']
        if lsda:
            return [' nspin=2']

        noncolin = kwargs['noncolin']
        if noncolin:
            return [' nspin=4']
        else:
            return [' nspin=1']
    except KeyError as err:
        logger.error("Missing required arguments when building "
                     "parameter '%s'! %s" % (name, err))
        return []


def set_ibrav_to_zero(name, **_kwargs):
    assert isinstance(name, str)
    line = ' ibrav=0'
    return [line]


def get_system_eamp(name, **kwargs):
    try:
        electric_potential = kwargs['electric_potential']
        if electric_potential in ('Berry_Phase', 'homogenous_field'):
            return []
        electric_field_amplitude = kwargs['electric_field_amplitude']
    except KeyError as err:
        logger.error("Missing required arguments when building "
                     "parameter '%s'! %s" % (name, err))
        return []

    if electric_potential == 'sawtooth_potential':
        return [' eamp={0}'.format(electric_field_amplitude)]
    else:
        return []


def get_electrons_efield(name, **kwargs):
    try:
        electric_potential = kwargs['electric_potential']
        if electric_potential in ('Berry_Phase', 'sawtooth_potential'):
            return []
        electric_field_amplitude = kwargs['electric_field_amplitude']
    except KeyError as err:
        logger.error("Missing required arguments when building "
                     "parameter '%s'! %s" % (name, err))
        return []

    if electric_potential == 'homogenous_field':
        return [' efield={0}'.format(electric_field_amplitude)]
    else:
        return []


def get_system_edir(name, **kwargs):
    try:
        electric_potential = kwargs['electric_potential']
        electric_field_direction = kwargs['electric_field_direction']
    except KeyError as err:
        logger.error("Missing required arguments when building "
                     "parameter '%s'! %s" % (name, err))
        return []

    if electric_potential == 'sawtooth_potential':
        return [' edir={0}'.format(electric_field_direction)]
    else:
        return []


def get_electric_potential_related(name, **kwargs):
    try:
        electric_potential = kwargs['electric_potential']
    except KeyError as err:
        logger.error("Missing required arguments when building "
                     "parameter '%s'! %s" % (name, err))
        return []

    if name == 'tefield':
        return [' %s=%s' % (name, to_fortran(electric_potential == 'sawtooth_potential'))]
    elif name == 'lelfield':
        return [' %s=%s' % (name, to_fortran(electric_potential == 'homogenous_field'))]
    elif name == 'lberry':
        return [' %s=%s' % (name, to_fortran(electric_potential == 'Berry_Phase'))]
    return []


def get_control_gdir(name, **kwargs):
    try:
        electric_potential = kwargs['electric_potential']
        electric_field_direction = kwargs['electric_field_direction']
    except KeyError as err:
        logger.error("Missing required arguments when building "
                     "parameter '%s'! %s" % (name, err))
        return []

    if electric_potential in ('homogenous_field', 'Berry_Phase'):
        return [' gdir={0}'.format(electric_field_direction)]
    else:
        return []


def get_cell_dofree(name, **kwargs):
    assert isinstance(name, str)
    cell_dofree_str = " cell_dofree = '%s'"
    cell_dofree_all = 'all'
    ret = [cell_dofree_str % cell_dofree_all]
    map_data = {
        'fix_volume': 'shape',
        'fix_area': '2Dshape',
        'fix_xy': '2Dxy',
        'isotropic': 'volume',
        'cell_do_free': kwargs.get("cell_do_free")
    }

    if len(set(kwargs).intersection(map_data)) > 1:
        logger.error("only one of %s can be true" % ', '.join(map_data))
        return ret

    for key, val in map_data.items():
        if kwargs.get(key):
            return [cell_dofree_str % val]

    return ret


def neb_set_system_nat(name, **kwargs):
    """
    Extract SYSTEM[nat] from the first element of the list of atomic_structure

    :param name: Variable name
    :param kwargs: list of dictionaries each containing an atomic_structure element
    :return: list containing one string to be printed in system name list nat = nat_value
    """
    assert isinstance(name, str)
    try:
        images = kwargs['atomic_structure']
    except KeyError:
        logger.error('No atomic_structure element found !!!')
        return []

    try:
        nat_value = images[0]['@nat']
    except (KeyError, IndexError):
        logger.error("error reading nat value from atomic_structure !!!")
        return []

    return [' nat = {0}'.format(nat_value)]


def ha2ry(name, **kwargs):
    related_tag = kwargs['_related_tag']
    value = kwargs[related_tag] * 2.e0
    return [' {} = {:12.8f}'.format(name, value)]


def set_one_amass_line(name, **kwargs):
    lines = []
    try:
        node = kwargs['amass']
        value = float(node['$'])
        index = node['@atom']
        lines.append(' {}({})={:7.3f}'.format(name, index, value))
    except TypeError:
        for node in kwargs['amass']:
            value = float(node['$'])
            index = node['@atom']
            lines.append(' {}({})={:7.3f}'.format(name, index, value))
    return lines


def set_one_proj_line(name, **kwargs):
    """
    writes projlines for epw input
    """
    lines = []
    try:
        node = kwargs['proj']
        value = str(node['$'])
        index = node['@atom']
        lines.append(f" {name}({index})='{value}'")
    except TypeError:
        for node in kwargs['proj']:
            value = str(node['$'])
            index = node['@atom']
            lines.append(f" {name}({index})='{value}'")
    return lines


def set_wdata_lines(name, **kwargs):
    """
    writes wdata strings for epw input
    """
    res = []
    if kwargs.get('wannier_plot', False):
        res = [" wdata(1) = 'bands_plot = .true.'",
               " wdata(2) = 'begin kpoint_path'"]
        iline = 2
        points = kwargs['wannier_plot_list']
        points.sort(key=lambda s: s['@segment'])
        for ppt in points:
            iline += 1
            res.append(f" wdata({iline}) = '{ppt['$'].strip()}'")
        iline += 1
        res.append(f" wdata({iline}) = 'end kpoint_path'")
        if kwargs.get('wannier_plot_format', None):
            iline += 1
            plot_format = kwargs['wannier_plot_format'].strip()
            res.append(f" wdata{iline} = 'bands_plot_format = {plot_format}'")
        use_ws_distance = "T" if kwargs.get('use_ws', False) else "F"
        iline += 1
        res.append(f" wdata({iline}) = 'use_ws_distance = {use_ws_distance}'")
    return res


def set_lda_plus_u_flag(name, **kwargs):
    """
    writes SYSTEM[set_lda_plus_u_flag] for old format
    """
    dftu = kwargs.get('dftU', None)
    if dftu:
        if dftu.get('@new_format'):
            return []
    assert isinstance(name, str)
    lines = []
    related_tag = kwargs['_related_tag']
    related_data = kwargs[related_tag]

    for value in iter(related_data if isinstance(related_data, list) else [related_data]):
        if value.get('@label') != 'no Hubbard' and value['$'] > 0:
            lines.append(f" {name} = .true.")
            break
    return lines


def set_lda_plus_u_kind(name, **kwargs):
    """
    writes SYSTEM[lda_plus_u_kind] flag for old format
    """
    dftu = kwargs.get('dftU', None)
    if dftu:
        if dftu.get('@new_format', False):
            return []
    #
    assert isinstance(name, str)
    lines = []
    if kwargs.get('lda_plus_u_kind', None) is not None:
        lines.append(f" lda_plus_u_kind = {kwargs['lda_plus_u_kind']}")
    return lines


def set_u_projection_type(name, **kwargs):
    """
    writes U_projection_type for old Hubbard format
    """
    assert isinstance(name, str)
    dftu = kwargs.get('dftU', None)
    if dftu:
        if dftu.get('@new_format', False):
            return []
    lines = []
    related_tag = kwargs['_related_tag']
    related_data = kwargs[related_tag]
    lines.append(f" {related_tag} = '{related_data}'")
    return lines


def set_boolean_flag(name, **kwargs):
    assert isinstance(name, str)
    lines = []
    related_tag = kwargs['_related_tag']
    related_data = kwargs[related_tag]
    if related_data is True or isinstance(related_data, str) and \
            related_data.strip() in ['true', 'True', 'TRUE', '1']:
        lines.append(' %s = .true.' % related_tag)
    else:
        lines.append(' %s = .false.' % related_tag)
    return lines


def set_what_td_calculation(name, **kwargs):
    assert isinstance(name, str)
    return [kwargs['whatTD']]


def get_xspectra_c(name, comma=False, **kwargs):
    """
    Get the xspectra component of a vector and print it in the form
    vector(component)=value

    We may add comma=True to enable the comma - not necessary and need to be tested.
    """
    value = float(kwargs[name])
    lines = []
    lines.append(f" {name[:-1]}({name[-1]})={value}{',' if comma else ''}")
    return lines



#
# Copyright (c), 2021, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
"""
A collection of functions for reading different files and quantities.
"""
import numpy as np
from xml.etree import ElementTree

__all__ = ['read_pseudo_file']


def read_pseudo_file(xml_file):
    """
    Reads a pseudo-potential XML-like file in the QE UPF format (text), returning
    the content of each tag in a dictionary. The file is read in strings and
    completed with a root UPF tag when it lacks, to avoids an XML syntax error.

    TODO: add support for UPF-schema files
    """
    def iter_upf_file():
        """
        Creates an iterator over the lines of an UPF file,
        inserting the root <UPF> tag when missing.
        """
        with open(xml_file, 'r') as f:
            fake_root = None
            for line in f:
                if fake_root is not None:
                    line = line.replace('&input', '&amp;input')
                    line = line.replace('&inputp', '&amp;inputp')
                    yield line
                else:
                    line = line.strip()
                    if line.startswith("<UPF") and line[4] in ('>', ' '):
                        yield line
                        fake_root = False
                    elif line:
                        yield "<UPF>"
                        yield line
                        fake_root = True
        if fake_root is True:
            yield "</UPF>"

    pseudo = {}
    psroot = ElementTree.fromstringlist(list(iter_upf_file()))

    # PP_INFO
    try:
        pp_info = psroot.find('PP_INFO').text
    except AttributeError:
        pp_info = ""
    try:
        pp_input = psroot.find('PP_INFO/PP_INPUTFILE').text
    except AttributeError:
        pp_input = ""
    pseudo.update(PP_INFO=dict(INFO=pp_info, PP_INPUT=pp_input))

    # PP_HEADER
    pseudo.update(PP_HEADER=dict(psroot.find('PP_HEADER').attrib))

    # PP_MESH
    pp_mesh = dict(psroot.find('PP_MESH').attrib)
    pp_r = np.array([float(x) for x in psroot.find('PP_MESH/PP_R').text.split()])
    pp_rab = np.array([float(x) for x in psroot.find('PP_MESH/PP_RAB').text.split()])
    pp_mesh.update(PP_R=pp_r, PP_RAB=pp_rab)
    pseudo.update(PP_MESH=pp_mesh)

    # PP_LOCAL
    node = psroot.find('PP_LOCAL')
    if node is not None:
        pp_local = np.array([x for x in map(float, node.text.split())])
    else:
        pp_local = None
    pseudo.update(PP_LOCAL=pp_local)

    # PP_RHOATOM
    node = psroot.find('PP_RHOATOM')
    if node is not None:
        pp_rhoatom = np.array([v for v in map(float, node.text.split())])
    else:
        pp_rhoatom = None
    pseudo.update(PP_RHOATOM=pp_rhoatom)

    # PP_NONLOCAL
    node = psroot.find('PP_NONLOCAL')
    if node is not None:
        betas = list()
        dij = None
        pp_aug = None
        pp_q = None
        for el in node:
            if 'PP_BETA' in el.tag:
                beta = dict(el.attrib)
                val = np.array(x for x in map(float, el.text.split()))
                beta.update(beta=val)
                betas.append(beta)
            elif 'PP_DIJ' in el.tag:
                text = '\n'.join(el.text.strip().split('\n')[1:])
                dij = np.array([x for x in map(float, text.split())])
            elif 'PP_AUGMENTATION' in el.tag:
                pp_aug = dict(el.attrib)
                pp_qijl = list()
                pp_qij = list()
                for q in el:
                    if 'PP_QIJL' in q.tag:
                        qijl = dict(q.attrib)
                        val = np.array(x for x in map(float, q.text.split()))
                        qijl.update(qijl=val)
                        pp_qijl.append(qijl)
                    elif 'PP_QIJ' in q.tag:
                        qij = dict(q.attrib)
                        val = np.array(x for x in map(float, q.text.split()))
                        qij.update(qij=val)
                        pp_qij.append(qij)
                    elif q.tag == 'PP_Q':
                        pp_q = np.array(x for x in map(float, q.text.split()))
                pp_aug.update(PP_QIJL=pp_qijl, PP_QIJ=pp_qij, PP_Q=pp_q)
        pp_nonlocal = dict(PP_BETA=betas, PP_DIJ=dij, PP_AUGMENTATION=pp_aug)
    else:
        pp_nonlocal = None

    pseudo.update(PP_NONLOCAL=pp_nonlocal)
    return pseudo



#
# Copyright (c), 2021, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
import numpy as np
import h5py

__all__ = ['read_charge_file', 'get_wf_attributes', 'get_wavefunctions',
           'get_wfc_miller_indices']


def read_charge_file(filename):
    """
    Reads a PW charge file in HDF5 format.

    :param filename: the name of the HDF5 file to read.
    :return: a dictionary describing the content of file \
    keys=[nr, ngm_g, gamma_only, rhog_, MillerIndexes]
    """
    with h5py.File(filename, "r") as h5f:
        MI = h5f.get('MillerIndices')[:]
        nr1 = 2 * max(abs(MI[:, 0])) + 1
        nr2 = 2 * max(abs(MI[:, 1])) + 1
        nr3 = 2 * max(abs(MI[:, 2])) + 1
        nr = np.array([nr1, nr2, nr3])
        res = dict(h5f.attrs.items())
        res.update({'MillInd': MI, 'nr_min': nr})
        rhog = h5f['rhotot_g'][:].reshape(res['ngm_g'], 2).dot([1.e0, 1.e0j])
        res['rhotot_g'] = rhog
        if 'rhodiff_g' in h5f.keys():
            rhog = h5f['rhodiff_g'][:].reshape(res['ngm_g'], 2).dot([1.e0, 1.e0j])
            res['rhodiff_g'] = rhog
        return res


# TODO update to the new format
def get_wf_attributes(filename):
    """
    Read attributes from a wfc HDF5 file.

    :param filename: the path to the wfc file
    :return: a dictionary with all attributes included reciprocal vectors
    """
    with h5py.File(filename, "r") as f:
        res = dict(f.attrs)
        mi_attrs = f.get('MillerIndices').attrs
        bg = np.array(mi_attrs.get(x) for x in ['bg1', 'bg2', 'bg3'])
        res.update({'bg': bg})
    return res


def get_wavefunctions(filename, start_band=None, stop_band=None):
    """
    Returns a numpy array with the wave functions for bands from start_band to
    stop_band. If not specified starts from 1st band and ends with last one.
    Band numbering is Python style starts from 0.abs

    :param filename: path to the wfc file
    :param start_band: first band to read, default first band in the file
    :param stop_band:  last band to read, default last band in the file
    :return: a numpy array with shape [nbnd,npw]
    """
    with h5py.File(filename, "r") as f:
        igwx = f.attrs.get('igwx')
        if start_band is None:
            start_band = 0
        if stop_band is None:
            stop_band = f.attrs.get('nbnd')
        if stop_band == start_band:
            stop_band = start_band + 1
        res = f.get('evc')[start_band:stop_band, :]

    res = np.asarray(x.reshape([igwx, 2]).dot([1.e0, 1.e0j]) for x in res[:])
    return res


def get_wfc_miller_indices(filename):
    """
    Reads miller indices from the wfc file

    :param filename: path to the wfc HDF5 file
    :return: a np.array of integers with shape [igwx,3]
    """
    with h5py.File(filename, "r") as f:
        res = f.get("MillerIndices")[:, :]
    return res
